@@ -2,6 +2,7 @@ use super::util;
 use ansi_term::Colour;
 use anyhow::{anyhow, Context, Result};
 use difference::{Changeset, Difference};
+use indoc::indoc;
 use lazy_static::lazy_static;
 use regex::bytes::{Regex as ByteRegex, RegexBuilder as ByteRegexBuilder};
 use regex::Regex;
@@ -48,7 +49,7 @@ pub enum TestEntry {
 
 impl Default for TestEntry {
     fn default() -> Self {
-        TestEntry::Group {
+        Self::Group {
             name: String::new(),
             children: Vec::new(),
             file_path: None,
@@ -56,82 +57,99 @@ impl Default for TestEntry {
     }
 }
 
-pub fn run_tests_at_path(
-    parser: &mut Parser,
-    path: &Path,
-    debug: bool,
-    debug_graph: bool,
-    filter: Option<&str>,
-    update: bool,
-) -> Result<()> {
-    let test_entry = parse_tests(path)?;
+pub struct TestOptions<'a> {
+    pub path: PathBuf,
+    pub debug: bool,
+    pub debug_graph: bool,
+    pub filter: Option<&'a str>,
+    pub include: Option<Regex>,
+    pub exclude: Option<Regex>,
+    pub update: bool,
+    pub open_log: bool,
+}
+
+pub fn run_tests_at_path(parser: &mut Parser, opts: &mut TestOptions) -> Result<()> {
+    let test_entry = parse_tests(&opts.path)?;
     let mut _log_session = None;
 
-    if debug_graph {
-        _log_session = Some(util::log_graphs(parser, "log.html")?);
-    } else if debug {
+    if opts.debug_graph {
+        _log_session = Some(util::log_graphs(parser, "log.html", opts.open_log)?);
+    } else if opts.debug {
         parser.set_logger(Some(Box::new(|log_type, message| {
             if log_type == LogType::Lex {
-                io::stderr().write(b"  ").unwrap();
+                io::stderr().write_all(b"  ").unwrap();
             }
-            write!(&mut io::stderr(), "{}\n", message).unwrap();
+            writeln!(&mut io::stderr(), "{message}").unwrap();
         })));
     }
 
     let mut failures = Vec::new();
     let mut corrected_entries = Vec::new();
+    let mut has_parse_errors = false;
     run_tests(
         parser,
         test_entry,
-        filter,
+        opts,
         0,
         &mut failures,
-        update,
         &mut corrected_entries,
+        &mut has_parse_errors,
     )?;
 
     parser.stop_printing_dot_graphs();
 
-    if failures.len() > 0 {
-        println!("");
+    if failures.is_empty() {
+        Ok(())
+    } else {
+        println!();
 
-        if update {
+        if opts.update && !has_parse_errors {
             if failures.len() == 1 {
-                println!("1 update:\n")
+                println!("1 update:\n");
             } else {
-                println!("{} updates:\n", failures.len())
+                println!("{} updates:\n", failures.len());
             }
 
             for (i, (name, ..)) in failures.iter().enumerate() {
-                println!("  {}. {}", i + 1, name);
+                println!("  {}. {name}", i + 1);
             }
+
             Ok(())
         } else {
-            if failures.len() == 1 {
-                println!("1 failure:")
-            } else {
-                println!("{} failures:", failures.len())
+            has_parse_errors = opts.update && has_parse_errors;
+
+            if !has_parse_errors {
+                if failures.len() == 1 {
+                    println!("1 failure:");
+                } else {
+                    println!("{} failures:", failures.len());
+                }
             }
 
             print_diff_key();
             for (i, (name, actual, expected)) in failures.iter().enumerate() {
-                println!("\n  {}. {}:", i + 1, name);
-                let actual = format_sexp_indented(&actual, 2);
-                let expected = format_sexp_indented(&expected, 2);
+                println!("\n  {}. {name}:", i + 1);
+                let actual = format_sexp_indented(actual, 2);
+                let expected = format_sexp_indented(expected, 2);
                 print_diff(&actual, &expected);
             }
-            Err(anyhow!(""))
+
+            if has_parse_errors {
+                Err(anyhow!(indoc! {"
+                    Some tests failed to parse with unexpected `ERROR` or `MISSING` nodes, as shown above, and cannot be updated automatically.
+                    Either fix the grammar or manually update the tests if this is expected."}))
+            } else {
+                Err(anyhow!(""))
+            }
         }
-    } else {
-        Ok(())
     }
 }
 
-pub fn check_queries_at_path(language: Language, path: &Path) -> Result<()> {
+pub fn check_queries_at_path(language: &Language, path: &Path) -> Result<()> {
     if path.exists() {
         for entry in WalkDir::new(path)
             .into_iter()
-            .filter_map(|e| e.ok())
+            .filter_map(std::result::Result::ok)
             .filter(|e| {
                 e.file_type().is_file()
                     && e.path().extension().and_then(OsStr::to_str) == Some("scm")
@@ -140,9 +158,9 @@ pub fn check_queries_at_path(language: Language, path: &Path) -> Result<()> {
         {
             let filepath = entry.file_name().to_str().unwrap_or("");
             let content = fs::read_to_string(entry.path())
-                .with_context(|| format!("Error reading query file {:?}", filepath))?;
+                .with_context(|| format!("Error reading query file {filepath:?}"))?;
             Query::new(language, &content)
-                .with_context(|| format!("Error in query file {:?}", filepath))?;
+                .with_context(|| format!("Error in query file {filepath:?}"))?;
         }
     }
     Ok(())
@@ -150,18 +168,18 @@ pub fn check_queries_at_path(language: Language, path: &Path) -> Result<()> {
 
 pub fn print_diff_key() {
     println!(
-        "\n{} / {}",
+        "\ncorrect / {} / {}",
         Colour::Green.paint("expected"),
-        Colour::Red.paint("actual")
+        Colour::Red.paint("unexpected")
     );
 }
 
-pub fn print_diff(actual: &String, expected: &String) {
+pub fn print_diff(actual: &str, expected: &str) {
     let changeset = Changeset::new(actual, expected, "\n");
     for diff in &changeset.diffs {
         match diff {
             Difference::Same(part) => {
-                print!("{}{}", part, changeset.split);
+                print!("{part}{}", changeset.split);
             }
             Difference::Add(part) => {
                 print!("{}{}", Colour::Green.paint(part), changeset.split);
@@ -171,17 +189,17 @@ pub fn print_diff(actual: &String, expected: &String) {
             }
         }
     }
-    println!("");
+    println!();
 }
 
 fn run_tests(
     parser: &mut Parser,
     test_entry: TestEntry,
-    filter: Option<&str>,
+    opts: &mut TestOptions,
     mut indent_level: i32,
     failures: &mut Vec<(String, String, String)>,
-    update: bool,
     corrected_entries: &mut Vec<(String, String, String, usize, usize)>,
+    has_parse_errors: &mut bool,
 ) -> Result<()> {
     match test_entry {
         TestEntry::Example {
@@ -192,31 +210,15 @@ fn run_tests(
             divider_delim_len,
             has_fields,
         } => {
-            if let Some(filter) = filter {
-                if !name.contains(filter) {
-                    if update {
-                        let input = String::from_utf8(input).unwrap();
-                        let output = format_sexp(&output);
-                        corrected_entries.push((
-                            name,
-                            input,
-                            output,
-                            header_delim_len,
-                            divider_delim_len,
-                        ));
-                    }
-                    return Ok(());
-                }
-            }
             let tree = parser.parse(&input, None).unwrap();
             let mut actual = tree.root_node().to_sexp();
             if !has_fields {
-                actual = strip_sexp_fields(actual);
+                actual = strip_sexp_fields(&actual);
             }
             print!("{}", "  ".repeat(indent_level as usize));
             if actual == output {
                 println!("✓ {}", Colour::Green.paint(&name));
-                if update {
+                if opts.update {
                     let input = String::from_utf8(input).unwrap();
                     let output = format_sexp(&output);
                     corrected_entries.push((
@@ -228,17 +230,35 @@ fn run_tests(
                     ));
                 }
             } else {
-                if update {
+                if opts.update {
                     let input = String::from_utf8(input).unwrap();
-                    let output = format_sexp(&actual);
-                    corrected_entries.push((
-                        name.clone(),
-                        input,
-                        output,
-                        header_delim_len,
-                        divider_delim_len,
-                    ));
-                    println!("✓ {}", Colour::Blue.paint(&name));
+                    let expected_output = format_sexp(&output);
+                    let actual_output = format_sexp(&actual);
+
+                    // Only bail early before updating if the actual is not the output, sometimes
+                    // users want to test cases that are intended to have errors, hence why this
+                    // check isn't shown above
+                    if actual.contains("ERROR") || actual.contains("MISSING") {
+                        *has_parse_errors = true;
+
+                        // keep the original `expected` output if the actual output has an error
+                        corrected_entries.push((
+                            name.clone(),
+                            input,
+                            expected_output,
+                            header_delim_len,
+                            divider_delim_len,
+                        ));
+                    } else {
+                        corrected_entries.push((
+                            name.clone(),
+                            input,
+                            actual_output,
+                            header_delim_len,
+                            divider_delim_len,
+                        ));
+                        println!("✓ {}", Colour::Blue.paint(&name));
+                    }
                 } else {
                     println!("✗ {}", Colour::Red.paint(&name));
                 }
@@ -247,12 +267,37 @@ fn run_tests(
         }
         TestEntry::Group {
             name,
-            children,
+            mut children,
             file_path,
         } => {
+            children.retain(|child| {
+                if let TestEntry::Example { name, .. } = child {
+                    if let Some(filter) = opts.filter {
+                        if !name.contains(filter) {
+                            return false;
+                        }
+                    }
+                    if let Some(include) = &opts.include {
+                        if !include.is_match(name) {
+                            return false;
+                        }
+                    }
+                    if let Some(exclude) = &opts.exclude {
+                        if exclude.is_match(name) {
+                            return false;
+                        }
+                    }
+                }
+                true
+            });
+
+            if children.is_empty() {
+                return Ok(());
+            }
+
             if indent_level > 0 {
                 print!("{}", "  ".repeat(indent_level as usize));
-                println!("{}:", name);
+                println!("{name}:");
             }
 
             let failure_count = failures.len();
@@ -262,16 +307,16 @@ fn run_tests(
                 run_tests(
                     parser,
                     child,
-                    filter,
+                    opts,
                     indent_level,
                     failures,
-                    update,
                     corrected_entries,
+                    has_parse_errors,
                 )?;
             }
 
             if let Some(file_path) = file_path {
-                if update && failures.len() - failure_count > 0 {
+                if opts.update && failures.len() - failure_count > 0 {
                     write_tests(&file_path, corrected_entries)?;
                 }
                 corrected_entries.clear();
@@ -281,18 +326,69 @@ fn run_tests(
     Ok(())
 }
 
-fn format_sexp(sexp: &String) -> String {
+fn format_sexp(sexp: &str) -> String {
     format_sexp_indented(sexp, 0)
 }
 
-fn format_sexp_indented(sexp: &String, initial_indent_level: u32) -> String {
+fn format_sexp_indented(sexp: &str, initial_indent_level: u32) -> String {
     let mut formatted = String::new();
 
+    if sexp.is_empty() {
+        return formatted;
+    }
+
     let mut indent_level = initial_indent_level;
     let mut has_field = false;
-    let mut s_iter = sexp.split(|c| c == ' ' || c == ')');
-    while let Some(s) = s_iter.next() {
-        if s.is_empty() {
+
+    let mut c_iter = sexp.chars().peekable();
+    let mut s = String::with_capacity(sexp.len());
+    let mut quote = '\0';
+    let mut saw_paren = false;
+    let mut did_last = false;
+
+    let mut fetch_next_str = |next: &mut String| {
+        next.clear();
+        while let Some(c) = c_iter.next() {
+            if c == '\'' || c == '"' {
+                quote = c;
+            } else if c == ' ' || (c == ')' && quote != '\0') {
+                if let Some(next_c) = c_iter.peek() {
+                    if *next_c == quote {
+                        next.push(c);
+                        next.push(*next_c);
+                        c_iter.next();
+                        quote = '\0';
+                        continue;
+                    }
+                }
+                break;
+            }
+            if c == ')' {
+                saw_paren = true;
+                break;
+            }
+            next.push(c);
+        }
+
+        // at the end
+        if c_iter.peek().is_none() && next.is_empty() {
+            if saw_paren {
+                // but did we see a ) before ending?
+                saw_paren = false;
+                return Some(());
+            }
+            if !did_last {
+                // but did we account for the end empty string as if we're splitting?
+                did_last = true;
+                return Some(());
+            }
+            return None;
+        }
+        Some(())
+    };
+
+    while fetch_next_str(&mut s).is_some() {
+        if s.is_empty() && indent_level > 0 {
             // ")"
             indent_level -= 1;
             write!(formatted, ")").unwrap();
@@ -301,7 +397,7 @@ fn format_sexp_indented(sexp: &String, initial_indent_level: u32) -> String {
                 has_field = false;
             } else {
                 if indent_level > 0 {
-                    writeln!(formatted, "").unwrap();
+                    writeln!(formatted).unwrap();
                     for _ in 0..indent_level {
                         write!(formatted, "  ").unwrap();
                     }
@@ -310,20 +406,27 @@ fn format_sexp_indented(sexp: &String, initial_indent_level: u32) -> String {
             }
 
             // "(node_name"
-            write!(formatted, "{}", s).unwrap();
+            write!(formatted, "{s}").unwrap();
 
             // "(MISSING node_name" or "(UNEXPECTED 'x'"
             if s.starts_with("(MISSING") || s.starts_with("(UNEXPECTED") {
-                let s = s_iter.next().unwrap();
-                write!(formatted, " {}", s).unwrap();
+                fetch_next_str(&mut s).unwrap();
+                if s.is_empty() {
+                    while indent_level > 0 {
+                        indent_level -= 1;
+                        write!(formatted, ")").unwrap();
+                    }
+                } else {
+                    write!(formatted, " {s}").unwrap();
+                }
             }
         } else if s.ends_with(':') {
             // "field:"
-            writeln!(formatted, "").unwrap();
+            writeln!(formatted).unwrap();
             for _ in 0..indent_level {
                 write!(formatted, "  ").unwrap();
             }
-            write!(formatted, "{} ", s).unwrap();
+            write!(formatted, "{s} ").unwrap();
             has_field = true;
             indent_level += 1;
         }
@@ -334,7 +437,7 @@ fn format_sexp_indented(sexp: &String, initial_indent_level: u32) -> String {
 
 fn write_tests(
     file_path: &Path,
-    corrected_entries: &Vec<(String, String, String, usize, usize)>,
+    corrected_entries: &[(String, String, String, usize, usize)],
 ) -> Result<()> {
     let mut buffer = fs::File::create(file_path)?;
     write_tests_to_buffer(&mut buffer, corrected_entries)
@@ -342,21 +445,19 @@ fn write_tests(
 
 fn write_tests_to_buffer(
     buffer: &mut impl Write,
-    corrected_entries: &Vec<(String, String, String, usize, usize)>,
+    corrected_entries: &[(String, String, String, usize, usize)],
 ) -> Result<()> {
     for (i, (name, input, output, header_delim_len, divider_delim_len)) in
         corrected_entries.iter().enumerate()
     {
         if i > 0 {
-            write!(buffer, "\n")?;
+            writeln!(buffer)?;
         }
         write!(
             buffer,
-            "{}\n{}\n{}\n{}\n{}\n\n{}\n",
+            "{}\n{name}\n{}\n{input}\n{}\n\n{}\n",
             "=".repeat(*header_delim_len),
-            name,
             "=".repeat(*header_delim_len),
-            input,
             "-".repeat(*divider_delim_len),
             output.trim()
         )?;
@@ -374,7 +475,7 @@ pub fn parse_tests(path: &Path) -> io::Result<TestEntry> {
         let mut children = Vec::new();
         for entry in fs::read_dir(path)? {
             let entry = entry?;
-            let hidden = entry.file_name().to_str().unwrap_or("").starts_with(".");
+            let hidden = entry.file_name().to_str().unwrap_or("").starts_with('.');
             if !hidden {
                 children.push(entry.path());
             }
@@ -382,7 +483,7 @@ pub fn parse_tests(path: &Path) -> io::Result<TestEntry> {
         children.sort_by(|a, b| {
             a.file_name()
                 .unwrap_or_default()
-                .cmp(&b.file_name().unwrap_or_default())
+                .cmp(b.file_name().unwrap_or_default())
         });
         let children = children
             .iter()
@@ -395,15 +496,16 @@ pub fn parse_tests(path: &Path) -> io::Result<TestEntry> {
         })
     } else {
         let content = fs::read_to_string(path)?;
-        Ok(parse_test_content(name, content, Some(path.to_path_buf())))
+        Ok(parse_test_content(name, &content, Some(path.to_path_buf())))
     }
 }
 
-pub fn strip_sexp_fields(sexp: String) -> String {
-    SEXP_FIELD_REGEX.replace_all(&sexp, " (").to_string()
+#[must_use]
+pub fn strip_sexp_fields(sexp: &str) -> String {
+    SEXP_FIELD_REGEX.replace_all(sexp, " (").to_string()
 }
 
-fn parse_test_content(name: String, content: String, file_path: Option<PathBuf>) -> TestEntry {
+fn parse_test_content(name: String, content: &str, file_path: Option<PathBuf>) -> TestEntry {
     let mut children = Vec::new();
     let bytes = content.as_bytes();
     let mut prev_name = String::new();
@@ -420,8 +522,8 @@ fn parse_test_content(name: String, content: String, file_path: Option<PathBuf>)
     // Find all of the `===` test headers, which contain the test names.
     // Ignore any matches whose suffix does not match the first header
     // suffix in the file.
-    let header_matches = HEADER_REGEX.captures_iter(&bytes).filter_map(|c| {
-        let header_delim_len = c.name("equals").map(|n| n.as_bytes().len()).unwrap_or(80);
+    let header_matches = HEADER_REGEX.captures_iter(bytes).filter_map(|c| {
+        let header_delim_len = c.name("equals").map_or(80, |m| m.as_bytes().len());
         let suffix1 = c
             .name("suffix1")
             .map(|m| String::from_utf8_lossy(m.as_bytes()));
@@ -433,8 +535,7 @@ fn parse_test_content(name: String, content: String, file_path: Option<PathBuf>)
             let test_name = c
                 .name("test_name")
                 .map(|c| String::from_utf8_lossy(c.as_bytes()).trim_end().to_string());
-            let res = Some((header_delim_len, header_range, test_name));
-            res
+            Some((header_delim_len, header_range, test_name))
         } else {
             None
         }
@@ -451,18 +552,16 @@ fn parse_test_content(name: String, content: String, file_path: Option<PathBuf>)
             let divider_range = DIVIDER_REGEX
                 .captures_iter(&bytes[prev_header_end..header_range.start])
                 .filter_map(|m| {
-                    let divider_delim_len =
-                        m.name("hyphens").map(|m| m.as_bytes().len()).unwrap_or(80);
+                    let divider_delim_len = m.name("hyphens").map_or(80, |m| m.as_bytes().len());
                     let suffix = m
                         .name("suffix")
                         .map(|m| String::from_utf8_lossy(m.as_bytes()));
                     if suffix == first_suffix {
                         let range = m.get(0).unwrap().range();
-                        let res = Some((
+                        Some((
                             divider_delim_len,
                             (prev_header_end + range.start)..(prev_header_end + range.end),
-                        ));
-                        res
+                        ))
                     } else {
                         None
                     }
@@ -520,7 +619,7 @@ mod tests {
     fn test_parse_test_content_simple() {
         let entry = parse_test_content(
             "the-filename".to_string(),
-            r#"
+            r"
 ===============
 The first test
 ===============
@@ -538,9 +637,8 @@ The second test
 d
 ---
 (d)
-        "#
-            .trim()
-            .to_string(),
+        "
+            .trim(),
             None,
         );
 
@@ -575,7 +673,7 @@ d
     fn test_parse_test_content_with_dashes_in_source_code() {
         let entry = parse_test_content(
             "the-filename".to_string(),
-            r#"
+            r"
 ==================
 Code with dashes
 ==================
@@ -596,9 +694,8 @@ abc
 -------------------
 
 (c (d))
-        "#
-            .trim()
-            .to_string(),
+        "
+            .trim(),
             None,
         );
 
@@ -631,9 +728,10 @@ abc
 
     #[test]
     fn test_format_sexp() {
+        assert_eq!(format_sexp(""), "");
         assert_eq!(
-            format_sexp(&"(a b: (c) (d) e: (f (g (h (MISSING i)))))".to_string()),
-            r#"
+            format_sexp("(a b: (c) (d) e: (f (g (h (MISSING i)))))"),
+            r"
 (a
   b: (c)
   (d)
@@ -641,18 +739,39 @@ abc
     (g
       (h
         (MISSING i)))))
+"
+            .trim()
+        );
+        assert_eq!(format_sexp("()"), "()");
+        assert_eq!(format_sexp("(A (M (B)))"), "(A\n  (M\n    (B)))");
+        assert_eq!(format_sexp("(A (U (B)))"), "(A\n  (U\n    (B)))");
+        assert_eq!(
+            format_sexp("(program (ERROR (UNEXPECTED ' ')) (identifier))"),
+            r"
+(program
+  (ERROR
+    (UNEXPECTED ' '))
+  (identifier))
+"
+            .trim()
+        );
+        assert_eq!(
+            format_sexp(r#"(source_file (MISSING ")"))"#),
+            r#"
+(source_file
+  (MISSING ")"))
+        "#
+            .trim()
+        );
+        assert_eq!(
+            format_sexp(r#"(source_file (ERROR (UNEXPECTED 'f') (UNEXPECTED '+')))"#),
+            r#"
+(source_file
+  (ERROR
+    (UNEXPECTED 'f')
+    (UNEXPECTED '+')))
 "#
             .trim()
-            .to_string()
-        );
-        assert_eq!(format_sexp(&"()".to_string()), "()".to_string());
-        assert_eq!(
-            format_sexp(&"(A (M (B)))".to_string()),
-            "(A\n  (M\n    (B)))"
-        );
-        assert_eq!(
-            format_sexp(&"(A (U (B)))".to_string()),
-            "(A\n  (U\n    (B)))"
         );
     }
 
@@ -678,7 +797,7 @@ abc
         write_tests_to_buffer(&mut buffer, &corrected_entries).unwrap();
         assert_eq!(
             String::from_utf8(buffer).unwrap(),
-            r#"
+            r"
 ================================================================================
 title 1
 ================================================================================
@@ -694,7 +813,7 @@ input 2
 --------------------------------------------------------------------------------
 
 output 2
-"#
+"
             .trim_start()
             .to_string()
         );
@@ -735,8 +854,7 @@ code
 
 (MISSING ";")
         "#
-            .trim()
-            .to_string(),
+            .trim(),
             None,
         );
 
@@ -779,7 +897,7 @@ code
     fn test_parse_test_content_with_suffixes() {
         let entry = parse_test_content(
             "the-filename".to_string(),
-            r#"
+            r"
 ==================asdf\()[]|{}*+?^$.-
 First test
 ==================asdf\()[]|{}*+?^$.-
@@ -818,9 +936,8 @@ NOT A TEST HEADER
 ---asdf\()[]|{}*+?^$.-
 
 (a)
-        "#
-            .trim()
-            .to_string(),
+        "
+            .trim(),
             None,
         );
 
@@ -853,7 +970,7 @@ NOT A TEST HEADER
                     },
                     TestEntry::Example {
                         name: "Test name with = symbol".to_string(),
-                        input: expected_input.clone(),
+                        input: expected_input,
                         output: "(a)".to_string(),
                         header_delim_len: 25,
                         divider_delim_len: 3,
@@ -869,7 +986,7 @@ NOT A TEST HEADER
     fn test_parse_test_content_with_newlines_in_test_names() {
         let entry = parse_test_content(
             "the-filename".to_string(),
-            r#"
+            r"
 ===============
 name
 with
@@ -885,8 +1002,7 @@ name with === signs
 code with ----
 ---
 (d)
-"#
-            .to_string(),
+",
             None,
         );
 
